MODULE DfModelEdit;

  IMPORT SYSTEM, Kernel, Files, Math, Log, Strings, DfConfig, DfUtils, DfRandom, DswSDL, DswSDLImage, DswGL;

  CONST
    (* anim *)
    stand = 0; walk = 1; die1 = 2; die2 = 3; attack = 4;
    seeup = 5; seedown = 6; attackup = 7; attackdown = 8; pain = 9;
    maxAnim = 10;

    (* weapons *)
    csaw = 0; hgun = 1; sg = 2; ssg = 3; mgun = 4; rkt = 5; plz = 6; bfg = 7;
    spl = 8; flm = 9;
    maxWeap = 10;

    (* directions *)
    left = 1; right = 0;
    maxDir = 2;

    (* weap directions *)
    forward = 0; up = 1; down = 2;
    maxWDir = 3;

    (* weapon states *)
    nofire = 0; fire = 1;
    maxWStates = 2;

    angleUp = -0.9599; (* -55 deg *)
    angleDown = 0.6108; (* 35 deg *)

    flipv = 0; fliph = 1;
    tps = 36;

    (* colors *)
    Tr = 0.0; Tg = 0.0; Tb = 0.0; Ta = 1.0; (* text *)
    Lr = 0.0; Lg = 1.0; Lb = 0.0; La = 0.5; (* current frame marker *)
    Pr = 1.0; Pg = 0.0; Pb = 0.0; Pa = 0.5; (* previous frame marker *)
    Sr = 0.0; Sg = 0.0; Sb = 1.0; Sa = 0.5; (* saved frame marker *)
    Hr = 1.0; Hg = 1.0; Hb = 1.0; Ha = 1.0; (* hit box *)

  TYPE
    GLHook = POINTER TO RECORD (DswGL.LoaderHook) END;
    EventHandler = RECORD (DswSDL.EventHandler) END;

    Point = RECORD
      x, y: INTEGER;
    END;

    Rect = RECORD
      x, y, w, h: INTEGER;
    END;

    Texture = POINTER TO RECORD
      id: DswGL.GLuint;
      w, h: INTEGER;
    END;

  VAR
    animName: ARRAY maxAnim OF DfConfig.Name;
    weapName: ARRAY maxWeap OF DfConfig.Name;
    weapBase: ARRAY maxWeap OF Point;
    flagBase: Point;
    player: Rect;
    corpse: Rect;

  VAR
    conf: DfConfig.Config;
    flagPoint: Point;
    flagAngle: INTEGER;
    flagTex: Texture;
    anim: ARRAY maxAnim OF RECORD
      name, mask: ARRAY maxDir OF Files.Name;
      a, m: ARRAY maxDir OF Texture;
      frames: INTEGER;
      back: BOOLEAN;
      wait: INTEGER;
      weap: ARRAY maxWeap OF RECORD
        points: POINTER TO ARRAY OF Point;
      END;
    END;
    weap: ARRAY maxWeap, maxWDir, maxWStates OF RECORD
      name: Files.Name;
      tex: Texture;
    END;
    font: Texture;

  VAR
    flagMode: BOOLEAN;
    curAnim, curFrame, curWeap: INTEGER;
    mdlScale, textScale, curDelay: INTEGER;
    Mr, Mg, Mb: REAL; (* mask color *)
    Br, Bg, Bb: REAL; (* background color *)
    wx, wy: INTEGER; (* model on screen offset *)
    mx, my: REAL; (* mouse *)
    motion: BOOLEAN;
    animate: BOOLEAN;

  VAR
    window: DswSDL.Window;
    context: DswSDL.GLContext;
    srcPath: Files.Name;
    quit: BOOLEAN;

  PROCEDURE (gl: GLHook) GetAddress (name: ARRAY OF SHORTCHAR): INTEGER;
  BEGIN
    RETURN DswSDL.hook.GLGetProcAddress(name)
  END GetAddress;

  PROCEDURE UpdateConfig;
    VAR i, j: INTEGER; s: POINTER TO ARRAY OF CHAR; p: ARRAY 1 OF Point; ss: ARRAY 32 OF CHAR;

    PROCEDURE AddPointsToString (IN p: ARRAY OF Point);
      VAR i, len: INTEGER; f, g: ARRAY 32 OF CHAR;
    BEGIN
      len := LEN(p) * 32 * 2 + 1;
      IF (s = NIL) OR (LEN(s) < len) THEN NEW(s, len) END;
      s^ := ""; len := LEN(p) - 1;
      FOR i := 0 TO len DO
        Strings.IntToString(p[i].x, f);
        Strings.IntToString(p[i].y, g);
        s^ := s$ + f + ":" + g;
        IF i # len THEN s^ := s$ + "," END;
      END
    END AddPointsToString;

  BEGIN
    FOR i := 0 TO maxAnim - 1 DO
      FOR j := 0 TO maxWeap - 1 DO
        IF anim[i].weap[j].points # NIL THEN
          AddPointsToString(anim[i].weap[j].points);
          conf.SetString(animName[i], weapName[j] + "_points", s)
        END
      END
    END;
    p[0] := flagPoint;
    AddPointsToString(p);
    conf.SetString("Model", "flag_point", s);
    Strings.IntToString(flagAngle, ss);
    conf.SetString("Model", "flag_angle", ss)
  END UpdateConfig;

  PROCEDURE WriteConfig (c: DfConfig.Config; IN path: ARRAY OF CHAR);
    CONST ln = 0DX + 0AX; (* CR+LN *)
    VAR loc: Files.Locator; f: Files.File; w: Files.Writer; dir, name: Files.Name;
    VAR s: DfConfig.Section; k: DfConfig.Key; res: INTEGER; ss: ARRAY 32 OF CHAR;
    VAR p: POINTER TO ARRAY OF SHORTCHAR;

    PROCEDURE WriteString (IN str: ARRAY OF CHAR);
      VAR len, res: INTEGER; p: POINTER TO ARRAY OF SHORTCHAR;
    BEGIN
      len := LEN(str$);
      IF len > 0 THEN
        IF (p = NIL) OR (LEN(p) < len * 3 + 1) THEN
          NEW(p, len * 3 + 1)
        END;
        Kernel.StringToUtf8(str, p, res);
        ASSERT(res = 0, 100);
        len := LEN(p$);
        w.WriteBytes(SYSTEM.THISARRAY(SYSTEM.ADR(p[0]), len), 0, len)
      END
    END WriteString;

  BEGIN
    ASSERT(c # NIL, 20);
    DfUtils.SplitPath(path, dir, name);
    loc := Files.dir.This(dir);
    f := Files.dir.New(loc, Files.dontAsk);
    IF f # NIL THEN
      w := f.NewWriter(NIL);
      s := c.sections;
      WHILE s # NIL DO
        WriteString("["); WriteString(s.name); WriteString("]" + ln);
        k := s.keys;
        WHILE k # NIL DO
          WriteString(k.name); WriteString("="); WriteString(k.value); WriteString(ln);
          k := k.next
        END;
        WriteString(ln);
        s := s.next
      END;
      f.Register(name, "txt", Files.dontAsk, res);
      IF res # 0 THEN
        Strings.IntToString(res, ss);
        Log.ParamMsg("error: unable to save config (^0)", ss, "", ""); Log.Ln
      END
    END
  END WriteConfig;

  PROCEDURE LoadBitmap (IN file: ARRAY OF CHAR; OUT rgba: POINTER TO ARRAY OF INTEGER; OUT w, h: INTEGER);
    VAR s: DswSDL.Surface; i, j: INTEGER;
  BEGIN
    rgba := NIL; w := 0; h := 0;
    s := DswSDLImage.Load(file);
    IF s # NIL THEN
      s := s.ConvertSurfaceFormat(DswSDL.ABGR8888, {}); (* !!! *)
      IF s # NIL THEN
        s.GetSize(w, h);
        NEW(rgba, w * h);
        FOR j := 0 TO h - 1 DO
          FOR i := 0 TO w - 1 DO
            rgba[j * w + i] := s.GetPixel(i, j)
          END
        END
      END
    END
  END LoadBitmap;

  PROCEDURE CreateTexture (IN rgba: ARRAY OF INTEGER; w, h: INTEGER): Texture;
    VAR tex: Texture; id: DswGL.GLuint;
  BEGIN tex := NIL; id := 0;
    DswGL.glGenTextures(1, id);
    IF id # 0 THEN
      NEW(tex); tex.id := id; tex.w := w; tex.h := h;
      DswGL.glBindTexture(DswGL.GL_TEXTURE_2D, id);
      DswGL.glTexParameterf(DswGL.GL_TEXTURE_2D, DswGL.GL_TEXTURE_MIN_FILTER, DswGL.GL_NEAREST);
      DswGL.glTexParameterf(DswGL.GL_TEXTURE_2D, DswGL.GL_TEXTURE_MAG_FILTER, DswGL.GL_NEAREST);
      DswGL.glTexParameterf(DswGL.GL_TEXTURE_2D, DswGL.GL_TEXTURE_WRAP_S, DswGL.GL_MIRRORED_REPEAT);
      DswGL.glTexParameterf(DswGL.GL_TEXTURE_2D, DswGL.GL_TEXTURE_WRAP_T, DswGL.GL_MIRRORED_REPEAT);
      DswGL.glTexImage2D(DswGL.GL_TEXTURE_2D, 0, DswGL.GL_RGBA, w, h, 0, DswGL.GL_RGBA, DswGL.GL_UNSIGNED_BYTE, SYSTEM.ADR(rgba[0]));
      DswGL.glBindTexture(DswGL.GL_TEXTURE_2D, 0)
    END;
    RETURN tex
  END CreateTexture;

  PROCEDURE LoadTexture (IN filename: ARRAY OF CHAR): Texture;
    VAR rgba: POINTER TO ARRAY OF INTEGER; w, h: INTEGER; tex: Texture;
  BEGIN
    rgba := NIL; w := 0; h := 0; tex := NIL;
    LoadBitmap(filename, rgba, w, h);
    IF rgba = NIL THEN LoadBitmap(filename + ".png", rgba, w, h) END;
    IF rgba = NIL THEN LoadBitmap(filename + ".tga", rgba, w, h) END;
    IF rgba # NIL THEN tex := CreateTexture(rgba, w, h) END;
    IF tex = NIL THEN Log.ParamMsg("unable to load texture '^1'", "", filename, ""); Log.Ln END;
    RETURN tex;
  END LoadTexture;

  PROCEDURE (tex: Texture) FINALIZE-;
    VAR id: DswGL.GLuint;
  BEGIN
    id := tex.id; tex.id := 0;
    IF id # 0 THEN DswGL.glDeleteTextures(1, id) END
  END FINALIZE;

  PROCEDURE DrawTexture (tex: Texture; x, y, fx, fy, fw, fh, sw, sh, angle: INTEGER; cr, cg, cb, ca: REAL; flags: SET);
    VAR tx, ty, tw, th, ffx, ffy: REAL;
  BEGIN
    ASSERT(tex # NIL, 20);
    ASSERT(fx >= 0, 21);
    ASSERT(fy >= 0, 22);
    ASSERT(fw >= 0, 23);
    ASSERT(fh >= 0, 24);
    ASSERT(sw >= 0, 25);
    ASSERT(sh >= 0, 26);
    tx := 1 / tex.w * fx;
    ty := 1 / tex.h * fy;
    tw := 1 / tex.w * fw;
    th := 1 / tex.h * fh;
    DswGL.glEnable(DswGL.GL_BLEND);
    DswGL.glEnable(DswGL.GL_TEXTURE_2D);
    DswGL.glBlendFunc(DswGL.GL_SRC_ALPHA, DswGL.GL_ONE_MINUS_SRC_ALPHA);
    DswGL.glBindTexture(DswGL.GL_TEXTURE_2D, tex.id);
    DswGL.glColor4d(cr, cg, cb, ca);
    DswGL.glPushMatrix;
    DswGL.glTranslated(x + fw * sw / 2, y + fh * sh / 2, 0);
    DswGL.glRotated(angle, 0, 0, 1);
    DswGL.glScaled(sw, sh, 1);
    IF flipv IN flags THEN ffx := -1 ELSE ffx := 1 END;
    IF fliph IN flags THEN ffy := -1 ELSE ffy := 1 END;
    DswGL.glBegin(DswGL.GL_QUADS);
    DswGL.glTexCoord2d(tx,      ty);      DswGL.glVertex2d(ffx * (-fw / 2), ffy * (-fh / 2));
    DswGL.glTexCoord2d(tx + tw, ty);      DswGL.glVertex2d(ffx * fw / 2,    ffy * (-fh / 2));
    DswGL.glTexCoord2d(tx + tw, ty + th); DswGL.glVertex2d(ffx * fw / 2,    ffy * fh / 2);
    DswGL.glTexCoord2d(tx,      ty + th); DswGL.glVertex2d(ffx * (-fw / 2), ffy * fh / 2);
    DswGL.glEnd;
    DswGL.glPopMatrix;
    DswGL.glBindTexture(DswGL.GL_TEXTURE_2D, 0);
    DswGL.glDisable(DswGL.GL_BLEND);
    DswGL.glDisable(DswGL.GL_TEXTURE_2D)
  END DrawTexture;

  PROCEDURE DrawText (IN str: ARRAY OF CHAR; x, y: INTEGER);
    VAR i, cx, cy, cw, ch: INTEGER;
  BEGIN
    i := 0; cw := font.w DIV 16; ch := font.h DIV 16;
    WHILE str[i] # 0X DO
      cx := ORD(str[i]) MOD 16 * cw; cy := ORD(str[i]) DIV 16 * ch;
      DrawTexture(font, x + i * cw * textScale, y, cx, cy, cw, ch, textScale, textScale, 0, Tr, Tg, Tb, Ta, {});
      INC(i)
    END
  END DrawText;

  PROCEDURE TextSize(IN str: ARRAY OF CHAR; OUT w, h: INTEGER);
  BEGIN
    w := LEN(str$) * (font.w DIV 16) * textScale;
    h := font.h DIV 16 * textScale
  END TextSize;

  PROCEDURE DrawLine (x0, y0, x1, y1, s: INTEGER; r, g, b, a: REAL);
  BEGIN
    DswGL.glEnable(DswGL.GL_BLEND);
    DswGL.glBlendFunc(DswGL.GL_SRC_ALPHA, DswGL.GL_ONE_MINUS_SRC_ALPHA);
    DswGL.glColor4d(r, g, b, a);
    DswGL.glLineWidth(s);
    DswGL.glBegin(DswGL.GL_LINES);
    DswGL.glVertex2d(x0, y0);
    DswGL.glVertex2d(x1, y1);
    DswGL.glEnd;
    DswGL.glDisable(DswGL.GL_BLEND)
  END DrawLine;

  PROCEDURE DrawRect (x0, y0, x1, y1, s: INTEGER; r, g, b, a: REAL);
  BEGIN
    DrawLine(x0, y0, x1, y0, s, r, g, b, a);
    DrawLine(x1, y0, x1, y1, s, r, g, b, a);
    DrawLine(x1, y1, x0, y1, s, r, g, b, a);
    DrawLine(x0, y1, x0, y0, s, r, g, b, a)
  END DrawRect;

  PROCEDURE LoadModelTextures;
    VAR i, j, k: INTEGER;
  BEGIN
    FOR i := 0 TO maxAnim - 1 DO
      IF anim[i].name[right] # "" THEN
        anim[i].a[right] := LoadTexture(srcPath + "/TEXTURES/" + anim[i].name[right])
      END;
      IF anim[i].name[left] # "" THEN
        anim[i].a[left] := LoadTexture(srcPath + "/TEXTURES/" + anim[i].name[left])
      END;
      IF anim[i].mask[right] # "" THEN
        anim[i].m[right] := LoadTexture(srcPath + "/TEXTURES/" + anim[i].mask[right])
      END;
      IF anim[i].mask[left] # "" THEN
        anim[i].m[left] := LoadTexture(srcPath + "/TEXTURES/" + anim[i].mask[left])
      END;
    END;
    FOR i := 0 TO maxWeap - 1 DO
      FOR j := 0 TO maxWDir - 1 DO
        FOR k := 0 TO maxWStates - 1 DO
          weap[i, j, k].tex := LoadTexture("WEAPONS/" + weap[i, j, k].name)
        END
      END
    END;
    flagTex := LoadTexture("TEXTURES/FLAGBLUE")
  END LoadModelTextures;

  PROCEDURE LoadModelConfig;
    VAR loc: Files.Locator; t: DfUtils.Text; ok: BOOLEAN; c: DfConfig.Config; s: ARRAY 32 OF CHAR;
    VAR sec: DfConfig.Section; i, j, k, version: INTEGER; p: POINTER TO ARRAY OF Point;

    PROCEDURE ReadString (sec: DfConfig.Section; IN key: DfConfig.Name; OUT s: ARRAY OF CHAR; IN def: ARRAY OF CHAR);
      VAR v: DfConfig.Value;
    BEGIN
      s := def$;
      IF sec # NIL THEN
        v := sec.GetString(key);
        IF v # NIL THEN
          s := v$
        END
      END
    END ReadString;

    PROCEDURE ReadInteger (sec: DfConfig.Section; IN key: DfConfig.Name; OUT x: INTEGER; min, max, def: INTEGER);
    BEGIN
      x := def;
      IF sec # NIL THEN
        x := sec.GetInt(key, min, max, def)
      END
    END ReadInteger;

    PROCEDURE ReadBoolean (sec: DfConfig.Section; IN key: DfConfig.Name; OUT x: BOOLEAN; def: BOOLEAN);
      VAR y: INTEGER;
    BEGIN
      IF def THEN y := -1 ELSE y := 0 END;
      ReadInteger(sec, key, y, MIN(INTEGER), MAX(INTEGER), y);
      x := y # 0
    END ReadBoolean;

    PROCEDURE ReadPoints (sec: DfConfig.Section; IN key: DfConfig.Name; OUT x: POINTER TO ARRAY OF Point; count: INTEGER);
      VAR v: DfConfig.Value; s: ARRAY 32 OF CHAR; pos, n, l: INTEGER;

      PROCEDURE ParsePoint (VAR p: Point);
        VAR s: ARRAY 32 OF CHAR; i, res: INTEGER;
      BEGIN
        i := 0; p.x := 0; p.y := 0;
        WHILE (v[pos] # 0X) & (v[pos] # ":") DO s[i] := v[pos]; INC(i); INC(pos) END;
        s[i] := 0X;
        Strings.StringToInt(s, p.x, res);
        ASSERT(res = 0, 100);
        IF v[pos] = ":" THEN
          INC(pos); i := 0;
          WHILE (v[pos] # 0X) & (v[pos] # ",") DO s[i] := v[pos]; INC(i); INC(pos) END;
          s[i] := 0X;
          Strings.StringToInt(s, p.y, res);
          ASSERT(res = 0, 101);
        END
      END ParsePoint;

    BEGIN x := NIL;
      ASSERT(count >= 0, 20);
      IF count > 0 THEN
        NEW(x, count);
        IF sec # NIL THEN
          v := sec.GetString(key);
          IF v # NIL THEN
            pos := 0; n := 0;
            ParsePoint(x[n]); INC(n);
            WHILE (n < count) & (v[pos] = ",") DO
              INC(pos); ParsePoint(x[n]); INC(n);
            END;
            IF n # count THEN
              Strings.IntToString(pos, s);
              Log.ParamMsg("error: [^0:^1:^2] ", sec.name, key, s);
              IF n > count THEN Log.String("too many points"); Log.Ln
              ELSIF n < count THEN
                IF v[pos] # 0X THEN Log.String("syntax error in points definition")
                ELSE Log.String("too few points")
                END;
                Log.Ln
              END
            END
          END
        END
      END
    END ReadPoints;

  BEGIN ok := TRUE;
    loc := Files.dir.This(srcPath).This("TEXT");
    t := DfUtils.ReadText(loc, "MODEL.txt");
    IF t # NIL THEN
      c := DfConfig.Parse(t);
      IF c # NIL THEN
        conf := c;
        sec := conf.FindSection("Model");
        IF sec # NIL THEN
          ReadInteger(sec, "version", version, MIN(INTEGER), MAX(INTEGER), 1);
          IF version = 1 THEN
            ReadPoints(sec, "flag_point", p, 1);
            IF p # NIL THEN flagPoint := p[0]
            ELSE flagPoint.x := 32; flagPoint.y := 16
            END;
            ReadInteger(sec, "flag_angle", flagAngle, MIN(INTEGER), MAX(INTEGER), -20);
            FOR i := 0 TO maxAnim - 1 DO
              sec := conf.FindSection(animName[i]);
              ReadString(sec, "resource", anim[i].name[right], "");
              ReadString(sec, "resource2", anim[i].name[left], "");
              ReadString(sec, "mask", anim[i].mask[right], "");
              ReadString(sec, "mask2", anim[i].mask[left], "");
              ReadInteger(sec, "frames", anim[i].frames, 0, MAX(INTEGER), 0);
              ReadBoolean(sec, "backanim", anim[i].back, FALSE);
              ReadInteger(sec, "waitcount", anim[i].wait, 1, MAX(INTEGER), 1);
              anim[i].wait := MAX(1, anim[i].wait DIV 3);
              FOR j := 0 TO maxWeap - 1 DO
                (* !!! set default values to flamethrower if points not specified *)
                IF ~(i IN {die1, die2, pain}) THEN
                  ReadPoints(sec, weapName[j] + "_points", anim[i].weap[j].points, anim[i].frames)
                END
              END
            END
          ELSE Log.String("unsupported model format version"); Log.Ln
          END
        ELSE Log.String("not model config"); Log.Ln
        END
      ELSE Log.String("error while parsing MODEL.txt (pos ="); Log.Int(c.pos); Log.Char(")"); Log.Ln
      END
    ELSE
      Strings.IntToString(loc.res, s);
      Log.ParamMsg("unable to open MODEL.txt (loc.res ^0)", s, "", ""); Log.Ln
    END;
    FOR i := 0 TO maxWeap - 1 DO
      FOR j := 0 TO maxWDir - 1 DO
        FOR k := 0 TO maxWStates - 1 DO
          Strings.ToUpper(weapName[i], weap[i, j, k].name);
          CASE j OF
          | forward: (* ignore *)
          | down: weap[i, j ,k].name := weap[i, j, k].name + "_DN"
          | up: weap[i, j, k].name := weap[i, j, k].name + "_UP"
          END;
          CASE k OF
          | nofire: (* ignore *)
          | fire: weap[i, j, k].name := weap[i, j, k].name + "_FIRE"
          END
        END
      END
    END;
    LoadModelTextures;
    curFrame := MIN(curFrame, anim[curAnim].frames)
  END LoadModelConfig;

  PROCEDURE Setup;
    VAR w, h: INTEGER;
  BEGIN
    window.GLGetDrawableSize(w, h);
    DswGL.glViewport(0, 0, w, h);
    DswGL.glMatrixMode(DswGL.GL_PROJECTION);
    DswGL.glLoadIdentity;
    DswGL.glOrtho(0, w, h, 0, -1, 1);
    DswGL.glMatrixMode(DswGL.GL_MODELVIEW);
    DswGL.glLoadIdentity;
  END Setup;

  PROCEDURE DrawState;
    VAR s, s0, s1: ARRAY 256 OF CHAR; w, h, y, xx, yy: INTEGER; conv: BOOLEAN;
  BEGIN y := 0;
    TextSize("", w, h);
    Strings.IntToString(curAnim, s);
    Strings.IntToString(maxAnim - 1, s0);
    IF animate THEN s1 := "(auto)" ELSE s1 := "" END;
    DrawText("Anim: " + s + "/" + s0 + " " + s1, 0, y); INC(y, h);
    Strings.IntToString(curFrame, s);
    Strings.IntToString(anim[curAnim].frames - 1, s0);
    DrawText("Frame: " + s + "/" + s0, 0, y); INC(y, h);
    IF curWeap >= 0 THEN s := weapName[curWeap]$ ELSE s := "--" END;
    DrawText("Weap: " + s, 0, y); INC(y, h);
    (* -------------------------------- *)
    conv := FALSE;
    IF flagMode THEN
      xx := flagPoint.x;
      yy := flagPoint.y;
      conv := TRUE
    ELSIF (curWeap >= 0) & (anim[curAnim].weap[curWeap].points # NIL) THEN
      xx := anim[curAnim].weap[curWeap].points[curFrame].x;
      yy := anim[curAnim].weap[curWeap].points[curFrame].y;
      conv := TRUE
    END;
    s0 := "--"; s := "--";
    IF conv THEN
      Strings.IntToString(xx, s);
      Strings.IntToString(yy, s0)
    END;
    DrawText("Current point: " + s + ":" + s0, 0, y); INC(y, h);
  END DrawState;

  PROCEDURE Draw;
    CONST len = 42;
    VAR w, h, wdir, wstate, x, y, f: INTEGER; tex: Texture; tw, th: INTEGER; s, c: REAL;
  BEGIN
    window.GLGetDrawableSize(w, h);
    DswGL.glClear(DswGL.GL_COLOR_BUFFER_BIT);
    DswGL.glClearColor(SHORT(Br), SHORT(Bg), SHORT(Bb), 1.0);
    IF (curWeap >= 0) & (anim[curAnim].weap[curWeap].points # NIL) THEN
      Math.SinCos(angleUp, s, c);
      DrawLine(
        wx + (64 DIV 2) * mdlScale,
        wy + (64 DIV 2) * mdlScale,
        wx + (64 DIV 2 + SHORT(ENTIER(c * len))) * mdlScale,
        wy + (64 DIV 2 + SHORT(ENTIER(s * len))) * mdlScale,
        1, Hr, Hg, Hb, Ha
      );
      Math.SinCos(angleDown, s, c);
      DrawLine(
        wx + (64 DIV 2) * mdlScale,
        wy + (64 DIV 2) * mdlScale,
        wx + (64 DIV 2 + SHORT(ENTIER(c * len))) * mdlScale,
        wy + (64 DIV 2 + SHORT(ENTIER(s * len))) * mdlScale,
        1, Hr, Hg, Hb, Ha
      );
      DrawLine(
        wx + (64 DIV 2) * mdlScale,
        wy + (64 DIV 2) * mdlScale,
        wx + (64 DIV 2 + len) * mdlScale,
        wy + (64 DIV 2) * mdlScale,
        1, Hr, Hg, Hb, Ha
      )
    END;
    DrawRect(wx, wy, wx + 64 * mdlScale, wy + 64 * mdlScale, 1, Hr, Hg, Hb, Ha);
    IF flagMode THEN
      DrawTexture(flagTex, wx + (flagPoint.x - 64 + flagBase.x) * mdlScale, wy + (flagPoint.y - flagBase.y + 1) * mdlScale, 0, 0, 64, 64, mdlScale, mdlScale, flagAngle, 1.0, 1.0, 1.0, 1.0, {flipv})
    END;
    IF curAnim IN {die1, die2} THEN
      DrawRect(
        wx + corpse.x * mdlScale,
        wy + corpse.y * mdlScale,
        wx + (corpse.x + corpse.w) * mdlScale,
        wy + (corpse.y + corpse.h) * mdlScale,
        1, Hr, Hg, Hb, Ha
      )
    ELSE
      DrawRect(
        wx + player.x * mdlScale,
        wy + player.y * mdlScale,
        wx + (player.x + player.w) * mdlScale,
        wy + (player.y + player.h) * mdlScale,
        1, Hr, Hg, Hb, Ha
      )
    END;
    IF (curWeap >= 0) & (anim[curAnim].weap[curWeap].points # NIL) THEN
      IF curAnim IN {attack, attackup, attackdown} THEN wstate := fire ELSE wstate := nofire END;
      IF curAnim IN {attackup, seeup} THEN wdir := up
      ELSIF curAnim IN {attackdown, seedown} THEN wdir := down
      ELSE wdir := forward
      END;
      tex := weap[curWeap, wdir, wstate].tex;
      x := (anim[curAnim].weap[curWeap].points[curFrame].x - weapBase[curWeap].x) * mdlScale;
      y := (anim[curAnim].weap[curWeap].points[curFrame].y - weapBase[curWeap].y) * mdlScale;
      DrawTexture(tex, wx + x, wy + y, 0, 0, tex.w, tex.h, mdlScale, mdlScale, 0,   1.0, 1.0, 1.0, 1.0, {});
    END;
    DrawTexture(anim[curAnim].a[right], wx, wy, curFrame * 64, 0, 64, 64, mdlScale, mdlScale, 0, 1.0, 1.0, 1.0, 1.0, {});
    DrawTexture(anim[curAnim].m[right], wx, wy, curFrame * 64, 0, 64, 64, mdlScale, mdlScale, 0, Mr, Mg, Mb, 1.0, {});
    DrawState;
    IF animate THEN
      IF curDelay > anim[curAnim].wait THEN
        curFrame := (curFrame + 1) MOD anim[curAnim].frames;
        IF (curAnim IN {die1, die2}) & (curFrame = anim[curAnim].frames - 1) THEN
          curDelay := -tps * 3 (* delay after die animation *)
        ELSE
          curDelay := 0
        END;
      END;
      INC(curDelay)
    END;
    window.GLSwap
  END Draw;

  PROCEDURE GetPosition;
  BEGIN
    IF flagMode THEN
      mx := flagPoint.x;
      my := flagPoint.y
    ELSIF anim[curAnim].weap[curWeap].points # NIL THEN
      mx := anim[curAnim].weap[curWeap].points[curFrame].x;
      my := anim[curAnim].weap[curWeap].points[curFrame].y
    END
  END GetPosition;

  PROCEDURE SetPosition (x, y: REAL; rel: BOOLEAN);
  BEGIN
    IF rel THEN mx := mx + x; my := my + y
    ELSE mx := x - wx DIV mdlScale; my := y - wy DIV mdlScale
    END;
    IF flagMode THEN
      flagPoint.x := SHORT(ENTIER(mx));
      flagPoint.y := SHORT(ENTIER(my))
    ELSIF anim[curAnim].weap[curWeap].points # NIL THEN
      anim[curAnim].weap[curWeap].points[curFrame].x := SHORT(ENTIER(mx));
      anim[curAnim].weap[curWeap].points[curFrame].y := SHORT(ENTIER(my))
    END
  END SetPosition;

  PROCEDURE (VAR e: EventHandler) Handle (VAR ev: DswSDL.Event);
  BEGIN
    WITH ev: DswSDL.Quit DO quit := TRUE
    | ev: DswSDL.WindowSizeChanged DO Setup
    | ev: DswSDL.MouseButtonDown DO GetPosition; motion := FALSE;
    | ev: DswSDL.MouseButtonUp DO
      IF motion = FALSE THEN
        SetPosition(ev.x / mdlScale, ev.y / mdlScale, FALSE)
      END
    | ev: DswSDL.MouseMotion DO
      IF ev.state # {} THEN
        SetPosition(ev.xrel / mdlScale, ev.yrel / mdlScale, TRUE);
        motion := TRUE
      END
    | ev: DswSDL.KeyDown DO
      IF DswSDL.ModLshift IN ev.keysym.mod THEN
        CASE ev.keysym.scancode OF
        | DswSDL.Minus: textScale := MAX(1, textScale - 1)
        | DswSDL.Equals: textScale := MAX(1, textScale + 1)
        ELSE (* ignore *)
        END
      ELSIF DswSDL.ModLctrl IN ev.keysym.mod THEN
        CASE ev.keysym.scancode OF
        | DswSDL.Minus: mdlScale := MAX(1, mdlScale - 1)
        | DswSDL.Equals: mdlScale := MAX(1, mdlScale + 1)
        | DswSDL.R: LoadModelTextures
        | DswSDL.Y: (* generate points *)
        | DswSDL.S: UpdateConfig; WriteConfig(conf, srcPath + "/TEXT/MODEL.txt")
        | DswSDL.O: LoadModelConfig
        | DswSDL.Q: quit := TRUE
        ELSE (* ignore *)
        END
      ELSE
        CASE ev.keysym.scancode OF
        | DswSDL.Leftbracket: curAnim := (curAnim - 1) MOD maxAnim; curFrame := 0; curDelay := 0
        | DswSDL.Rightbracket: curAnim := (curAnim + 1) MOD maxAnim; curFrame := 0; curDelay := 0
        | DswSDL.Backspace: curFrame := (curFrame - 1) MOD anim[curAnim].frames; curDelay := 0
        | DswSDL.Return: curFrame := (curFrame + 1) MOD anim[curAnim].frames; curDelay := 0
        | DswSDL.Grave: flagMode := ~flagMode
        | DswSDL.Q: IF flagMode THEN flagAngle := (flagAngle - 1) MOD 360 END
        | DswSDL.E: IF flagMode THEN flagAngle := (flagAngle + 1) MOD 360 END
        | DswSDL.SCANCODE_1: curWeap := -1
        | DswSDL.SCANCODE_2: curWeap := csaw
        | DswSDL.SCANCODE_3: curWeap := hgun
        | DswSDL.SCANCODE_4: curWeap := sg
        | DswSDL.SCANCODE_5: curWeap := ssg
        | DswSDL.SCANCODE_6: curWeap := mgun
        | DswSDL.SCANCODE_7: curWeap := rkt
        | DswSDL.SCANCODE_8: curWeap := plz
        | DswSDL.SCANCODE_9: curWeap := bfg
        | DswSDL.SCANCODE_0: curWeap := spl
        | DswSDL.Minus: curWeap := flm
        | DswSDL.Space: animate := ~animate
        | DswSDL.Up, DswSDL.W, DswSDL.K: GetPosition; SetPosition(0, -1, TRUE)
        | DswSDL.Down, DswSDL.S, DswSDL.J: GetPosition; SetPosition(0, +1, TRUE)
        | DswSDL.Left, DswSDL.A, DswSDL.H: GetPosition; SetPosition(-1, 0, TRUE)
        | DswSDL.Right, DswSDL.D, DswSDL.L: GetPosition; SetPosition(+1, 0, TRUE)
        | DswSDL.R:
          Mr := DfRandom.Get() MOD 256 / 255;
          Mg := DfRandom.Get() MOD 256 / 255;
          Mb := DfRandom.Get() MOD 256 / 255
        | DswSDL.B:
          Br := DfRandom.Get() MOD 256 / 255;
          Bg := DfRandom.Get() MOD 256 / 255;
          Bb := DfRandom.Get() MOD 256 / 255
        ELSE (* ignore *)
        END
      END
    ELSE
    END
  END Handle;

  PROCEDURE Update;
    VAR w, h: INTEGER;
  BEGIN
    window.GLGetDrawableSize(w, h);
    wx := w DIV 2 - 64 * mdlScale DIV 2;
    wy := h DIV 2 - 64 * mdlScale DIV 2
  END Update;

  PROCEDURE Main;
    CONST
      title = "Doom 2D Forever Model Editor";
      wx = DswSDL.centered; wy = DswSDL.centered;
      ww = 640; wh = 480;
      wflags = {DswSDL.opengl, DswSDL.resizable, DswSDL.allowHighDpi};
    VAR
      ok: BOOLEAN; gl: GLHook; h: EventHandler;
  BEGIN
    IF Kernel.trapCount # 0 THEN Kernel.Quit(1) END;
    IF Kernel.argc = 2 THEN srcPath := Kernel.argv[1]$
    ELSE Log.ParamMsg("usage: ^0 <modeldir>", Kernel.argv[0]$, "", ""); Log.Ln; Kernel.Quit(1)
    END;
    ok := FALSE;
    IF DswSDL.GLSetAttribute(DswSDL.GL_CONTEXT_MAJOR_VERSION, 1) THEN
      IF DswSDL.GLSetAttribute(DswSDL.GL_CONTEXT_MINOR_VERSION, 1) THEN
        window := DswSDL.CreateWindow(title, wx, wy, ww, wh, wflags);
        IF window # NIL THEN
          context := window.GLCreateContext();
          IF context # NIL THEN
            NEW(gl); DswGL.SetLoaderHook(gl); DswGL.Init;
            ok := TRUE
          ELSE Log.ParamMsg("^0: can't create context", Kernel.argv[0]$, "", ""); Log.Ln
          END
        ELSE Log.ParamMsg("^0: can't create window", Kernel.argv[0]$, "", ""); Log.Ln
        END
      ELSE Log.ParamMsg("^0: can't set minor context version", Kernel.argv[0]$, "", ""); Log.Ln
      END
    ELSE Log.ParamMsg("^0: can't set major context version", Kernel.argv[0]$, "", ""); Log.Ln
    END;
    IF ~ok THEN Kernel.Quit(1) END;
    font := LoadTexture("default");
    IF font = NIL THEN
      Log.ParamMsg("^0: cant load font", Kernel.argv[0]$, "", ""); Log.Ln;
      Kernel.Quit(1)
    END;
    Setup;
    LoadModelConfig;
    WHILE ~quit DO
      WHILE DswSDL.HandleEvent(h) DO END;
      Update;
      Draw;
      DswSDL.Delay(1000 DIV tps)
    END;
    Kernel.Quit(0)
  END Main;

  PROCEDURE InitWeap (i: INTEGER; IN name: ARRAY OF CHAR; x, y: INTEGER);
  BEGIN
    weapName[i] := name$; weapBase[i].x := x; weapBase[i].y := y;
  END InitWeap;

  PROCEDURE Init;
  BEGIN
    mdlScale := 5;
    textScale := 2;
    animName[stand] := "StandAnim";
    animName[walk] := "WalkAnim";
    animName[die1] := "Die1Anim";
    animName[die2] := "Die2Anim";
    animName[attack] := "AttackAnim";
    animName[seeup] := "SeeUpAnim";
    animName[seedown] := "SeeDownAnim";
    animName[attackup] := "AttackUpAnim";
    animName[attackdown] := "AttackDownAnim";
    animName[pain] := "PainAnim";
    InitWeap(csaw, "CSaw", 8, 4);
    InitWeap(hgun, "HGun", 8, 8);
    InitWeap(sg, "Sg", 16, 16);
    InitWeap(ssg, "Ssg", 16, 24);
    InitWeap(mgun, "MGun", 16, 16);
    InitWeap(rkt, "Rkt", 24, 24);
    InitWeap(plz, "Plz", 16, 16);
    InitWeap(bfg, "BFG", 24, 24);
    InitWeap(spl, "Spl", 16, 16);
    InitWeap(flm, "Flm", 8, 8);
    flagBase.x := 16; flagBase.y := 43;
    player.x := 15; player.y := 12;
    player.w := 34; player.h := 52;
    corpse.x := 15; corpse.y := 48;
    corpse.w := 34; corpse.h := 16;
    Mr := DfRandom.Get() MOD 256 / 255;
    Mg := DfRandom.Get() MOD 256 / 255;
    Mb := DfRandom.Get() MOD 256 / 255;
    Br := DfRandom.Get() MOD 256 / 255;
    Bg := DfRandom.Get() MOD 256 / 255;
    Bb := DfRandom.Get() MOD 256 / 255;
  END Init;

BEGIN
  Init;
  Kernel.Start(Main)
END DfModelEdit.
