MODULE DswSDLLin32;

  IMPORT S := SYSTEM, Kernel, SDL := DswSDL, Lib := DswSDLLin32Lib;

  TYPE
    Hook = POINTER TO RECORD (SDL.Hook) END;

    Window = POINTER TO RECORD (SDL.Window)
      ptr: Lib.Window;
    END;

    GLContext = POINTER TO RECORD (SDL.GLContext)
      ptr: Lib.GLContext;
    END;
    
    WindowIdentifier = RECORD (Kernel.Identifier)
      ptr: Lib.Window;
    END;
    
    RWops* = POINTER TO LIMITED RECORD (SDL.RWops)
      base: SDL.RWops;
      ptr: Lib.RWops;
      res: Lib.Int;
    END;

    Surface = POINTER TO RECORD (SDL.Surface)
      ptr: Lib.Surface;
    END;

    Identifier = RECORD (Kernel.Identifier)
      ptr: Lib.GLContext;
    END;

    Events = RECORD [union]
      quit: SDL.Quit;
      key: SDL.KeyDown;  (* SDL.KeyboardEvent *)
      text: SDL.TextInput;
      motion: SDL.MouseMotion;
      button: SDL.MouseButtonUp;  (* SDL.MouseButtonEvent *)
      wheel: SDL.MouseWheel;
      window: SDL.WindowShown;  (* SDL.WindowEvent *)
    END;
    
  (* Surface *)

  PROCEDURE NewSurface* (ptr: Lib.Surface): Surface;
    VAR surf: Surface;
  BEGIN
    surf := NIL;
    IF ptr # NIL THEN
      NEW(surf); surf.ptr := ptr
    END;
    RETURN surf
  END NewSurface;

  PROCEDURE (s: Surface) GetSize* (VAR w, h: INTEGER);
  BEGIN
    w := s.ptr.w;
    h := s.ptr.h
  END GetSize;

  PROCEDURE (s: Surface) GetPixel* (x, y: INTEGER): INTEGER;
    VAR v1: ARRAY 3 OF BYTE; v2: SHORTINT; v4, i, bpp: INTEGER;
  BEGIN
    ASSERT(x >= 0, 20);
    ASSERT(y >= 0, 21);
    ASSERT(x < s.ptr.w, 22);
    ASSERT(y < s.ptr.h, 23);
    bpp := s.ptr.format.BytesPerPixel;
    i := y * s.ptr.pitch + x * bpp;
    CASE bpp OF
    | 1:
      S.GET(s.ptr.pixels + i, v1[0]);
      v4 := v1[0] MOD 256
    | 2:
      S.GET(s.ptr.pixels + i, v2);
      v4 := v2 MOD 65536
    | 3:
      S.GET(s.ptr.pixels + i, v1[0]);
      S.GET(s.ptr.pixels + i + 1, v1[1]);
      S.GET(s.ptr.pixels + i + 2, v1[2]);
      v4 := v1[0] MOD 256 * 65536 + v1[1] MOD 256 * 256 + v1[2] MOD 256
    | 4:
      S.GET(s.ptr.pixels + i, v4)
    END;
    RETURN v4
  END GetPixel;

  PROCEDURE (s: Surface) ConvertSurfaceFormat* (format: INTEGER; flags: SET): Surface;
    VAR surf: Lib.Surface;
  BEGIN
    surf := Lib.ConvertSurfaceFormat(s.ptr, format, ORD(flags));
    RETURN NewSurface(surf)
  END ConvertSurfaceFormat;

  PROCEDURE (surf: Surface) FINALIZE-;
  BEGIN
    IF (surf.ptr # NIL) & (surf.ptr.refcount > 0) THEN
      Lib.FreeSurface(surf.ptr);
      surf.ptr := NIL
    END
  END FINALIZE;

  (* Hook *)

  PROCEDURE (h: Hook) WasInit (flags: SET): SET;
  BEGIN
    RETURN BITS(Lib.WasInit(ORD(flags)))
  END WasInit;

  PROCEDURE (h0: Hook) CreateWindow (IN title: ARRAY OF CHAR; x, y, w, h: INTEGER; flags: SET): Window;
    VAR utf: POINTER TO ARRAY OF SHORTCHAR; ptr: Lib.Window; win: Window; res: INTEGER;
  BEGIN
    IF title = "" THEN
      ptr := Lib.CreateWindow("", x, y, w, h, ORD(flags))
    ELSE
      NEW(utf, LEN(title$) * 3 + 1);
      Kernel.StringToUtf8(title, utf, res);
      ptr := Lib.CreateWindow(utf, x, y, w, h, ORD(flags))
    END;
    IF ptr = NIL THEN win := NIL
    ELSE NEW(win); win.ptr := ptr
    END;
    RETURN win
  END CreateWindow;

  PROCEDURE (h: Hook) Delay (ms: INTEGER);
  BEGIN
    Lib.Delay(ms)
  END Delay;

  PROCEDURE (h: Hook) GetMouseState (OUT x, y: INTEGER): SET;
    VAR xx, yy: Lib.Int; buttons: SET;
  BEGIN
    buttons := Lib.GetMouseState(xx, yy);
(* !!!
    x := S.CONV(INTEGER, xx);
    y := S.CONV(INTEGER, yy);
*)
    x := S.VAL(INTEGER, xx);
    y := S.VAL(INTEGER, yy);
    RETURN buttons
  END GetMouseState;

  PROCEDURE (h: Hook) StartTextInput;
  BEGIN
    Lib.StartTextInput
  END StartTextInput;

  PROCEDURE (h: Hook) StopTextInput;
  BEGIN
    Lib.StopTextInput
  END StopTextInput;

  PROCEDURE (h: Hook) HandleEvent (VAR handler: SDL.EventHandler): BOOLEAN;
    VAR res, tag: INTEGER; ev: Lib.Event; events: Events;
  BEGIN
    res := Lib.PollEvent(ev);
    IF res # 0 THEN
      CASE ev.type OF
      | Lib.QUIT:
          events.quit.timestamp := ev.quit.timestamp;
          handler.Handle(events.quit)
      | Lib.KEYUP, Lib.KEYDOWN:
          events.key.timestamp := ev.key.timestamp;
          events.key.windowID := ev.key.windowID;
          events.key.state := ev.key.state;
          events.key.repeat := ev.key.repeat;
          events.key.keysym.scancode := ev.key.keysym.scancode;
          events.key.keysym.sym := ev.key.keysym.sym;
          events.key.keysym.mod := BITS(ev.key.keysym.mod);
          CASE ev.type OF
          | Lib.KEYUP: tag := S.TYP(SDL.KeyUp)
          | Lib.KEYDOWN: tag := S.TYP(SDL.KeyDown)
          END;
          handler.Handle(S.THISRECORD(S.ADR(events.key), tag))
      | Lib.TEXTINPUT:
          events.text.timestamp := ev.text.timestamp;
          events.text.windowID := ev.text.windowID;
          Kernel.Utf8ToString(ev.text.text, events.text.text, res);
          handler.Handle(events.text)
      | Lib.MOUSEBUTTONDOWN, Lib.MOUSEBUTTONUP:
          events.button.timestamp := ev.button.timestamp;
          events.button.windowID := ev.button.windowID;
          events.button.which := ev.button.which;
          events.button.button := ev.button.button - 1; (* <-- different interface *)
          events.button.state := BITS(ev.button.state);
          events.button.clicks := ev.button.clicks;
          events.button.x := ev.button.x;
          events.button.y := ev.button.y;
          CASE ev.type OF
          | Lib.MOUSEBUTTONDOWN: tag := S.TYP(SDL.MouseButtonDown)
          | Lib.MOUSEBUTTONUP: tag := S.TYP(SDL.MouseButtonUp)
          END;
          handler.Handle(S.THISRECORD(S.ADR(events.button), tag))
      | Lib.MOUSEMOTION:
          events.motion.timestamp := ev.motion.timestamp;
          events.motion.windowID := ev.motion.windowID;
          events.motion.which := ev.motion.which;
          events.motion.state := BITS(ev.motion.state);
          events.motion.x := ev.motion.x;
          events.motion.y := ev.motion.y;
          events.motion.xrel := ev.motion.xrel;
          events.motion.yrel := ev.motion.yrel;
          handler.Handle(events.motion)
      | Lib.MOUSEWHEEL:
          events.wheel.timestamp := ev.wheel.timestamp;
          events.wheel.windowID := ev.wheel.windowID;
          events.wheel.which := ev.wheel.which;
          events.wheel.x := ev.wheel.x;
          events.wheel.y := ev.wheel.y;
          events.wheel.direction := ev.wheel.direction;
          handler.Handle(events.wheel)
      | Lib.WINDOWEVENT:
          events.window.timestamp := ev.window.timestamp;
          events.window.windowID := ev.window.windowID;
          events.window.data1 := ev.window.data1;
          events.window.data2 := ev.window.data2;
          CASE ev.window.event OF
          | Lib.WINDOWEVENT_SHOWN: tag := S.TYP(SDL.WindowShown)
          | Lib.WINDOWEVENT_HIDDEN: tag := S.TYP(SDL.WindowHidden)
          | Lib.WINDOWEVENT_EXPOSED: tag := S.TYP(SDL.WindowExposed)
          | Lib.WINDOWEVENT_MOVED: tag := S.TYP(SDL.WindowMoved)
          | Lib.WINDOWEVENT_RESIZED: tag := S.TYP(SDL.WindowResized)
          | Lib.WINDOWEVENT_SIZE_CHANGED: tag := S.TYP(SDL.WindowSizeChanged)
          | Lib.WINDOWEVENT_MINIMIZED: tag := S.TYP(SDL.WindowMinimized)
          | Lib.WINDOWEVENT_MAXIMIZED: tag := S.TYP(SDL.WindowMaximized)
          | Lib.WINDOWEVENT_RESTORED: tag := S.TYP(SDL.WindowRestored)
          | Lib.WINDOWEVENT_ENTER: tag := S.TYP(SDL.WindowEnter)
          | Lib.WINDOWEVENT_LEAVE: tag := S.TYP(SDL.WindowLeave)
          | Lib.WINDOWEVENT_FOCUS_GAINED: tag := S.TYP(SDL.WindowFocusGained)
          | Lib.WINDOWEVENT_FOCUS_LOST: tag := S.TYP(SDL.WindowFocusLost)
          | Lib.WINDOWEVENT_CLOSE: tag := S.TYP(SDL.WindowClose)
          | Lib.WINDOWEVENT_TAKE_FOCUS: tag := S.TYP(SDL.WindowTakeFocus)
          | Lib.WINDOWEVENT_HIT_TEST: tag := S.TYP(SDL.WindowHitTest)
          END;
          handler.Handle(S.THISRECORD(S.ADR(events.window), tag))
      ELSE
        (* not implemented, just ignore *)
      END
    END;
    RETURN res # 0
  END HandleEvent;

  PROCEDURE [ccall16] RWSize (context: Lib.RWops): Lib.Sint64;
    VAR rw: SDL.RWops;
  BEGIN
    rw := S.VAL(SDL.RWops, context.hidden.unknown.data1);
    RETURN rw.Size()
  END RWSize;

  PROCEDURE [ccall16] RWSeek (context: Lib.RWops; offset: Lib.Sint64; whence: Lib.Int): Lib.Sint64;
    VAR rw: SDL.RWops;
  BEGIN
    rw := S.VAL(SDL.RWops, context.hidden.unknown.data1);
    RETURN rw.Seek(SHORT(offset), whence)
  END RWSeek;

  PROCEDURE [ccall16] RWRead (context: Lib.RWops; ptr: Lib.PtrVoid; size, maxnum: Lib.SizeT): Lib.SizeT;
    VAR rw: SDL.RWops;
  BEGIN
    rw := S.VAL(SDL.RWops, context.hidden.unknown.data1);
    RETURN rw.Read(S.THISARRAY(ptr, size * maxnum), size, maxnum)
  END RWRead;

  PROCEDURE [ccall16] RWWrite (context: Lib.RWops; ptr: Lib.PtrVoid; size, num: Lib.SizeT): Lib.SizeT;
    VAR rw: SDL.RWops;
  BEGIN
    ASSERT((ptr # 0) OR (size * num = 0), 100);
    rw := S.VAL(SDL.RWops, context.hidden.unknown.data1);
    RETURN rw.Write(S.THISARRAY(ptr, size * num), size, num)
  END RWWrite;

  PROCEDURE [ccall16] RWClose (context: Lib.RWops): Lib.Int;
    VAR rw: RWops;
  BEGIN
    rw := S.VAL(RWops, context.hidden.unknown.data2);
    rw.res := rw.base.Close();
    Lib.FreeRW(rw.ptr);
    rw.base := NIL;
    rw.ptr := NIL;
    RETURN rw.res
  END RWClose;

  PROCEDURE (h: Hook) RWFromUnknown (base: SDL.RWops): RWops;
    VAR rw: RWops; ptr: Lib.RWops;
  BEGIN
    ASSERT(base # NIL, 20);
    rw := NIL;
    ptr := Lib.AllocRW();
    IF ptr # NIL THEN
      NEW(rw); rw.base := base; rw.ptr := ptr;
      ptr.type := Lib.RWOPS_UNKNOWN;
      ptr.size := RWSize;
      ptr.seek := RWSeek;
      ptr.read := RWRead;
      ptr.write := RWWrite;
      ptr.close := RWClose;
      ptr.hidden.unknown.data1 := S.ADR(base^);
      ptr.hidden.unknown.data2 := S.ADR(rw^)
    END;
    RETURN rw
  END RWFromUnknown;
  
  PROCEDURE (h: Hook) LoadBMP (rw: SDL.RWops; freesrc: BOOLEAN): Surface;
    VAR ptr: Lib.Surface; free: Lib.Int;
  BEGIN
    ASSERT(rw # NIL, 20);
    ASSERT(~rw.Closed(), 21);
    WITH rw: RWops DO
      IF freesrc THEN free := Lib.True ELSE free := Lib.False END;
      ptr := Lib.LoadBMP_RW(rw.ptr, free);
      RETURN NewSurface(ptr);
    END
  END LoadBMP;

  PROCEDURE (h: Hook) GLLoadLibrary (IN path: ARRAY OF SHORTCHAR): BOOLEAN;
  BEGIN
    RETURN Lib.GL_LoadLibrary(path) = 0
  END GLLoadLibrary;

  PROCEDURE (h: Hook) GLGetProcAddress (IN proc: ARRAY OF SHORTCHAR): INTEGER;
  BEGIN
    RETURN Lib.GL_GetProcAddress(proc)
  END GLGetProcAddress;

  PROCEDURE (h: Hook) GLUnloadLibrary;
  BEGIN
    Lib.GL_UnloadLibrary
  END GLUnloadLibrary;

  PROCEDURE (h: Hook) GLExtensionSupported (IN extension: ARRAY OF SHORTCHAR): BOOLEAN;
  BEGIN
    RETURN Lib.GL_ExtensionSupported(extension) # 0
  END GLExtensionSupported;

  PROCEDURE (h: Hook) GLResetAttributes;
  BEGIN
    Lib.GL_ResetAttributes
  END GLResetAttributes;

  PROCEDURE (h: Hook) GLSetAttribute (attr, value: INTEGER): BOOLEAN;
  BEGIN
    RETURN Lib.GL_SetAttribute(attr, value) = 0
  END GLSetAttribute;

  PROCEDURE (h: Hook) GLGetAttribute (attr: INTEGER; VAR value: INTEGER): BOOLEAN;
  BEGIN
    RETURN Lib.GL_GetAttribute(attr, value) = 0
  END GLGetAttribute;

  PROCEDURE (VAR id: WindowIdentifier) Identified (): BOOLEAN;
  BEGIN
    RETURN id.obj(Window).ptr = id.ptr
  END Identified;

  PROCEDURE (h: Hook) GLGetCurrentWindow* (): Window;
    VAR id: WindowIdentifier; win: Window; ptr: Lib.Window; p: ANYPTR;
  BEGIN
    win := NIL;
    ptr := Lib.GL_GetCurrentWindow();
    IF ptr # NIL THEN
      id.ptr := ptr;
      id.typ := S.TYP(Window);
      p := Kernel.ThisFinObj(id);
      IF p # NIL THEN  (* founded in cache *)
        win := p(Window)
      ELSE
        NEW(win); win.ptr := ptr
      END
    END;
    RETURN win
  END GLGetCurrentWindow;

  PROCEDURE (VAR id: Identifier) Identified (): BOOLEAN;
  BEGIN
    RETURN id.obj(GLContext).ptr = id.ptr
  END Identified;

  PROCEDURE (h: Hook) GLGetCurrentContext (): GLContext;
    VAR id: Identifier; ptr: Lib.GLContext; gl: GLContext; p: ANYPTR;
  BEGIN
    gl := NIL;
    ptr := Lib.GL_GetCurrentContext();
    IF ptr # NIL THEN
      id.ptr := ptr;
      id.typ := S.TYP(GLContext);
      p := Kernel.ThisFinObj(id);
      IF p # NIL THEN  (* founded in cache *)
        gl := p(GLContext)
      ELSE
        NEW(gl); gl.ptr := ptr
      END
    END;
    RETURN gl
  END GLGetCurrentContext;

  PROCEDURE (h: Hook) GLSetSwapInterval (interval: INTEGER): BOOLEAN;
  BEGIN
    RETURN Lib.GL_SetSwapInterval(interval) = 0
  END GLSetSwapInterval;

  PROCEDURE (h: Hook) GLGetSwapInterval (): INTEGER;
  BEGIN
    RETURN Lib.GL_GetSwapInterval()
  END GLGetSwapInterval;

  (* Window *)

  PROCEDURE (win: Window) GetID (): INTEGER;
  BEGIN
    RETURN Lib.GetWindowID(win.ptr)
  END GetID;

  PROCEDURE (win: Window) SetInputFocus (): BOOLEAN;
  BEGIN
    RETURN Lib.SetWindowInputFocus(win.ptr) = 0
  END SetInputFocus;

  PROCEDURE (win: Window) Hide;
  BEGIN
    Lib.HideWindow(win.ptr)
  END Hide;

  PROCEDURE (win: Window) Show;
  BEGIN
    Lib.ShowWindow(win.ptr)
  END Show;

  PROCEDURE (win: Window) GLGetDrawableSize (OUT w, h: INTEGER);
  BEGIN
    ASSERT(win.ptr # NIL, 20);
    Lib.GL_GetDrawableSize(win.ptr, w, h)
  END GLGetDrawableSize;

  PROCEDURE (win: Window) GLCreateContext (): GLContext;
    VAR gl: GLContext; ptr: Lib.GLContext;
  BEGIN
    ASSERT(win.ptr # NIL, 20);
    ptr := Lib.GL_CreateContext(win.ptr);
    IF ptr = NIL THEN gl := NIL
    ELSE NEW(gl);gl.ptr := ptr
    END;
    RETURN gl
  END GLCreateContext;

  PROCEDURE (win: Window) GLMakeCurrent (gl: SDL.GLContext): BOOLEAN;
    VAR ptr: Lib.GLContext;
  BEGIN
    ASSERT(win.ptr # NIL, 20);
    IF gl = NIL THEN ptr := NIL
    ELSE
      WITH gl: GLContext DO
        ptr := gl.ptr;
        ASSERT(ptr # NIL, 21)
      END
    END;
    RETURN Lib.GL_MakeCurrent(win.ptr, ptr) = 0
  END GLMakeCurrent;

  PROCEDURE (win: Window) GLSwap;
  BEGIN
    ASSERT(win.ptr # NIL, 20);
    Lib.GL_SwapWindow(win.ptr)
  END GLSwap;

  PROCEDURE (win: Window) Destroy;
  BEGIN
    IF win.ptr # NIL THEN
      Lib.DestroyWindow(win.ptr);
      win.ptr := NIL;
    END
  END Destroy;

  PROCEDURE (win: Window) FINALIZE-;
    (* needed for caching *)
  END FINALIZE;
  
  (* GLContext *)

  PROCEDURE (gl: GLContext) Delete;
  BEGIN
    IF gl.ptr # NIL THEN
      Lib.GL_DeleteContext(gl.ptr);
      gl.ptr := NIL
    END
  END Delete;

  PROCEDURE (gl: GLContext) FINALIZE-;
    (* needed for caching *)
  END FINALIZE;

  (* RWops *)

  PROCEDURE (rw: RWops) Size* (): INTEGER;
  BEGIN
    ASSERT(rw.ptr # NIL, 20);
    RETURN SHORT(rw.ptr.size(rw.ptr))
  END Size;

  PROCEDURE (rw: RWops) Seek* (offset, whence: INTEGER): INTEGER;
  BEGIN
    ASSERT(rw.ptr # NIL, 20);
    RETURN SHORT(rw.ptr.seek(rw.ptr, offset, whence))
  END Seek;

  PROCEDURE (rw: RWops) Read* (VAR ptr: ARRAY OF BYTE; size, maxnum: INTEGER): INTEGER;
  BEGIN
    ASSERT(rw.ptr # NIL, 20);
    RETURN SHORT(rw.ptr.read(rw.ptr, S.ADR(ptr[0]), size, maxnum))
  END Read;

  PROCEDURE (rw: RWops) Write* (IN ptr: ARRAY OF BYTE; size, num: INTEGER): INTEGER;
  BEGIN
    ASSERT(rw.ptr # NIL, 20);
    RETURN SHORT(rw.ptr.write(rw.ptr, S.ADR(ptr[0]), size, num))
  END Write;

  PROCEDURE (rw: RWops) Close* (): INTEGER;
  BEGIN
    IF rw.ptr # NIL THEN
      rw.res := rw.ptr.close(rw.ptr);
      rw.base := NIL;
      rw.ptr := NIL
    END;
    RETURN rw.res
  END Close;

  PROCEDURE (rw: RWops) Closed* (): BOOLEAN;
  BEGIN
    RETURN rw.ptr = NIL
  END Closed;

  PROCEDURE (rw: RWops) GetPtr* (): Lib.RWops, NEW;
  BEGIN
    RETURN rw.ptr
  END GetPtr;

  PROCEDURE (rw: RWops) FINALIZE-;
    VAR res: INTEGER;
  BEGIN
    res := rw.Close()
  END FINALIZE;

  (* init *)

  PROCEDURE Init;
    VAR res: Lib.Int; hook: Hook;
  BEGIN
    res := Lib.Init(ORD(SDL.everything));
    NEW(hook); SDL.SetHook(hook)
  END Init;

  PROCEDURE Fin;
  BEGIN
    Lib.Quit
  END Fin;

BEGIN
  Init
CLOSE
  Fin
END DswSDLLin32.
